;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(defsystem clio
    (:pretty-name "Common Lisp Interactive Objects")
  (:module clue clue (:type :system))
  (:parallel
   clue
   (:parallel
    (:forms :compile 
	    (progn
	      (setq *compile-system-proclamation* *normal-declaration*)
	      (proclaim *compile-system-proclamation*)
	      #+akcl (setq compiler::*compile-ordinaries* t)))
    "clio"
    "exports"
    (:load "precom")
    "ol-defs"
    "utility"
    "core-mixins"
    "gravity"
    "buffer"
    "text-command"
    "display-text"
    "ol-images"
    "buttons"
    "confirm"
    "scroller"
    "table"
    "choices"
    "form"
    "menu"
    "psheet"
    "command"
    "edit-text"
    "slider"
    "scroll-frame"
    "mchoices"
    "dlog-button"
    "display-imag"
    (:compile "precom")
    )))
