;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(set-system-source-file 'clio-examples
			(subfile '("clue" "clio" "examples") :name "sysdef"))

;(pushnew 'clio-examples *auto-load-systems*)
