;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(set-system-source-file 'clio (subfile '("clue" "clio") :name "sysdef"))

;(pushnew 'clio *auto-load-systems*)

