;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(defsystem clio-examples
    (:pretty-name "CLIO Example Programs")
  (:module clio clio (:type :system))
  (:parallel
   clio
   (:parallel
    (:forms :compile
	    (progn
	      (setq *compile-system-proclamation* *normal-declaration*)
	      (proclaim *compile-system-proclamation*)
	      #+akcl (setq compiler::*compile-ordinaries* t)))
    "package"
    "exports"
    (:load "precom")
    "cmd-frame"
    "sketchpad"
    "sketch"
    (:compile "precom"))))


